# Ruby grammar for Racc based on Matz parse.y from Ruby 1.6.5
# Original author: Yukihiro Matsumoto
# "Rubified" by:   Robert Feldt
#
# Original file is Copyright (C) 1993-2001 Yukihiro Matsumoto
# This file is Copyright (C) 2001 Robert Feldt
#

class Parser
 token kCLASS kMODULE kDEF kUNDEF kBEGIN kRESCUE kENSURE kEND kIF kUNLESS
 token kTHEN kELSIF kELSE kCASE kWHEN kWHILE kUNTIL kFOR kBREAK kNEXT kREDO
 token kRETRY kIN kDO kDO_COND kDO_BLOCK kRETURN kYIELD kSUPER kSELF kNIL
 token kTRUE kFALSE kAND kOR kNOT kIF_MOD kUNLESS_MOD kWHILE_MOD kUNTIL_MOD
 token kRESCUE_MOD kALIAS kDEFINED klBEGIN klEND k__LINE__ k__FILE__

 token tIDENTIFIER tFID tGVAR tSVAR tIVAR tCONSTANT tCVAR
 token tINTEGER tFLOAT tSTRING tXSTRING tREGEXP
 token tDSTRING tDXSTRING tDREGEXP tNTH_REF tBACK_REF

 token tUPLUS                # unary+ 
 token tUMINUS               # unary- 
 token tPOW                  # ** 
 token tCMP                  # <=> 
 token tEQ                   # == 
 token tEQQ                  # === 
 token tNEQ                  # != 
 token tGEQ                  # >= 
 token tLEQ                  # <= 
 token tANDOP tOROP          # && and || 
 token tMATCH tNMATCH        # =~ and !~ 
 token tDOT2 tDOT3           # .. and ... 
 token tAREF tASET           # [] and []= 
 token tLSHFT tRSHFT         # << and >> 
 token tCOLON2               # :: 
 token tCOLON3               # :: at EXPR_BEG 
 token tOP_ASGN              # +=, -=  etc. 
 token tASSOC                # => 
 token tLPAREN               # ( 
 token tLPAREN_ARG           # ( 
 token tRPAREN               # ) 
 token tLBRACK               # [ 
 token tLBRACE               # { 
 token tLBRACE_ARG           # { 
 token tSTAR                 # * 
 token tAMPER                # & 
 token tSYMBEG

preclow
  left  kIF_MOD kUNLESS_MOD kWHILE_MOD kUNTIL_MOD kRESCUE_MOD
  left  kOR kAND
  right kNOT
  nonassoc kDEFINED
  right '=' tOP_ASGN
  right '?' ':'
  nonassoc tDOT2 tDOT3
  left  tOROP
  left  tANDOP
  nonassoc  tCMP tEQ tEQQ tNEQ tMATCH tNMATCH
  left  '>' tGEQ '<' tLEQ
  left  '|' '^'
  left  '&'
  left  tLSHFT tRSHFT
  left  '+' '-'
  left  '*' '/' '%'
  right '!' '~' tUPLUS tUMINUS
  right tPOW
prechigh

options no_result_var

rule

program         :  {
                     # val[0] = ???
                     @lexer.state = EXPR_BEG
                     # top_local_init()
                     # if ((VALUE)ruby_class == rb_cObject) 
		     #   @class_nest = 0
                     # else 
		     #   @class_nest = 1
		     # end
                   }
                  compstmt
                    {
                      # Lots of stuff here I deleted!
		      take_action(:on_compound_statement, val.indices(1))
                    }

compstmt        : stmts opt_terms
                    {
                      # void_stmts(val[0])
                      val[0] # No action since we only drop opt_terms
                    }

stmts           : none  
                    { nil } 
                | stmt  
                    { take_action(:on_single_statement, val) }
                | stmts terms stmt
                    { take_action(:on_another_statement, val) }
                | error stmt
                    { take_action(:on_statements_error, val) }

stmt            : kALIAS fitem {@lexer.state = EXPR_FNAME} fitem
                    {
		      error_unless_alias_valid
		      take_action(:on_alias, val.indices(0,1,3))
                    }
                | kALIAS tGVAR tGVAR
                    {
		      error_unless_alias_valid
		      take_action(:on_alias, val)
                    }
                | kALIAS tGVAR tSVAR
                    {
		      error_unless_alias_valid
		      take_action(:on_alias, val)
                    }
                | kALIAS tGVAR tBACK_REF
                    {
		      error_unless_alias_valid
		      take_action(:on_alias, val)
                    }
                | kALIAS tGVAR tNTH_REF
                    {
		      yyerror("can't make alias for the number variables")
		      nil
                    }
                | kUNDEF undef_list
                    {
		      if (@in_def > 0) || @in_single
		        yyerror("undef within method") 
		      end
		      take_action(:on_undef, val)
                    }
                | stmt kIF_MOD expr
                    {
		      # TODO: Should warn if expr is String or Regexp, see
		      # cond in parse.y.
		      take_action(:on_if_mod, val)
                    }
                | stmt kUNLESS_MOD expr
                    {
		      # TODO: Should warn if expr is String or Regexp, see
		      # cond in parse.y.
		      take_action(:on_unless_mod, val)
                    }
                | stmt kWHILE_MOD expr
                    {
		      # TODO: Should warn if expr is String or Regexp, see
		      # cond in parse.y.
		      take_action(:on_while_mod, val)
                    }
                | stmt kUNTIL_MOD expr
                    {
		      # TODO: Should warn if expr is String or Regexp, see
		      # cond in parse.y.
		      take_action(:on_until_mod, val)
                    }
                | stmt kRESCUE_MOD stmt
                    { take_action(:on_rescue_mod, val) }
                | klBEGIN
                    {
		      if (@in_def>0 || @in_single)
                        raise ParseError, "BEGIN in method"
                      end
                    }
                  '{' compstmt '}'
                    { take_action(:on_begin_block, val.indices(0, 2, 3, 4)) }
                | klEND '{' compstmt '}'
                    {
		      if (@compile_for_eval && (@in_def>0 || @in_single))
		        raise ParseError, "BEGIN in method"
		      end
		      take_action(:on_end_block, val) 
		    }
                | lhs '=' command_call
                    {
		      # TODO Should check that val[2] is a value expression
		      take_action(:on_assignment, val)
                    }
                | mlhs '=' command_call
                    {
		      # TODO Should check that val[2] is a value expression
		      take_action(:on_assignment, val)
                    }
                | lhs '=' mrhs_basic
                    { take_action(:on_assignment, val) }
                | mlhs '=' mrhs
                    {
		      # TODO Should check that val[2] is a value expression
		      take_action(:on_assignment, val)
                    }
                | expr

expr            : kRETURN call_args
                    {
		      if (!@compile_for_eval && (!(@in_def>0)) && !@in_single)
		        yyerror("return appeared outside of method")
		      end
                      # TODO: Check for return argument validity!
		      take_action(:on_return, val)
                    }
                | kBREAK call_args
                    {
                      # TODO: Check for return argument validity!
		      take_action(:on_break, val)
                    }
                | kNEXT call_args
                    {
                      # TODO: Check for return argument validity!
		      take_action(:on_next, val)
                    }
                | command_call
                | expr kAND expr
                    { take_action(:on_and, val) }
                | expr kOR expr
                    { take_action(:on_or, val) }
                | kNOT expr
                    {
		      # Should check that expr is valid conditional and value
		      take_action(:on_not, val)
                    }
                | '!' command_call
                    {
		      # Should check that expr is valid conditional and value
		      take_action(:on_exclamation, val)
                    }
                | arg

command_call    : command
                | block_command

block_command   : block_call
                | block_call '.' operation2 command_args
                    { take_action(:on_call, val) }
                | block_call tCOLON2 operation2 command_args
                    { take_action(:on_call, val) }

command         : operation command_args
                    { take_action(:on_call, ([nil]*2) + val) }
                | primary '.' operation2 command_args
                    { take_action(:on_call, val) }
                | primary tCOLON2 operation2 command_args
                    { take_action(:on_call, val) }
                | kSUPER command_args
                    {
                      if (!@compile_for_eval && !(@in_def>0) && !@in_single)
		        yyerror("super called outside of method") 
		      end
		      take_action(:on_super, val)
                    }
                | kYIELD call_args
                    {
		      # Should check that no block argument given!
		      take_action(:on_yield, val)
                    }

# Hmm, I find these strange! Why parens both at mlhs and mlhs_entry?
# Check with matz?

mlhs            : mlhs_basic
                | tLPAREN mlhs_entry ')' # Can this one be deleted?
                    { val[1] }

mlhs_entry      : mlhs_basic
                | tLPAREN mlhs_entry ')'
                    { val[1] }

mlhs_basic      : mlhs_head
                    { take_action(:on_mlhs, val) }
                | mlhs_head mlhs_item
                    { take_action(:on_mlhs, val[0].push(val[1])) }
                | mlhs_head tSTAR mlhs_node
                    { take_action(:on_mlhs, [val[0], val[2]]) }
                | mlhs_head tSTAR
                    { take_action(:on_mlhs, [val[0], true]) }
                | tSTAR mlhs_node
                    { take_action(:on_mlhs, [nil, val[1]]) }
                | tSTAR
                    { take_action(:on_mlhs, [nil, true]) }

mlhs_item       : mlhs_node
                | tLPAREN mlhs_entry ')'
                    { val[1] }

mlhs_head       : mlhs_item ','
                    { [val[0]] }
                | mlhs_head mlhs_item ','
                    { val[0].push(val[1]) }

mlhs_node       : variable
                    {
		      # Should check that val[0] is assignable
		      val[0]
                    }
                | primary '[' aref_args ']'
                    { take_action(:on_slice_var, val) }
                | primary '.' tIDENTIFIER
                    { take_action(:on_access, val) }
                | primary tCOLON2 tIDENTIFIER
                    { take_action(:on_access, val) }
                | primary '.' tCONSTANT
                    { take_action(:on_access, val) }
                | backref
                    {
		      backref_error(val[0])
		      nil
                    }

lhs             : variable
                    {
		      # Check that val[0] assignable
		      val[0]
                    }
                | primary '[' aref_args ']'
                    { take_action(:on_slice_var, val) }
                | primary '.' tIDENTIFIER
                    { take_action(:on_access, val) }
                | primary tCOLON2 tIDENTIFIER
                    { take_action(:on_access, val) }
                | primary '.' tCONSTANT
                    { take_action(:on_access, val) }
                | backref
                    {
		      backref_error(val[0])
		      nil
                    }

cname           : tIDENTIFIER
                    { yyerror("class/module name must be CONSTANT") }
                | tCONSTANT

fname           : tIDENTIFIER
                | tCONSTANT
                | tFID
                | op
                    {
		      @lexer.state = EXPR_END
		      val[0]
                    }
                | reswords
                    {
		      @lexer.state = EXPR_END
		      val[0]
                    }

fitem           : fname
                | symbol

undef_list      : fitem
                    { val }
                | undef_list ',' {@lexer.state = EXPR_FNAME} fitem
                    { val[0].push(val[3]) }

op              : '|'
                | '^'
                | '&'
                | tCMP
                | tEQ
                | tEQQ
                | tMATCH
                | '>'
                | tGEQ
                | '<'
                | tLEQ
                | tLSHFT
                | tRSHFT
                | '+'
                | '-'
                | '*'
                | tSTAR
                | '/'
                | '%'
                | tPOW
                | '~'
                | tUPLUS
                | tUMINUS
                | tAREF
                | tASET
                | '`'

reswords        : k__LINE__ | k__FILE__  | klBEGIN | klEND
                | kALIAS | kAND | kBEGIN | kBREAK | kCASE | kCLASS | kDEF
                | kDEFINED | kDO | kELSE | kELSIF | kEND | kENSURE | kFALSE
                | kFOR | kIF_MOD | kIN | kMODULE | kNEXT | kNIL | kNOT
                | kOR | kREDO | kRESCUE | kRETRY | kRETURN | kSELF | kSUPER
                | kTHEN | kTRUE | kUNDEF | kUNLESS_MOD | kUNTIL_MOD | kWHEN
                | kWHILE_MOD | kYIELD | kRESCUE_MOD

arg             : lhs '=' arg
                    {
		      # Should check that val[2] is valid value expr
		      take_action(:on_assignment, val)
                    }
                | variable tOP_ASGN arg
                    {
		      # Should check that val[0] is assignable
		      # Should check that val[0] is gettable
		      take_action(:on_op_assign, val)
		    }
                | primary '[' aref_args ']' tOP_ASGN arg
                    {
		      dest = take_action(:on_slice_var, [val[0]] + val[2])
		      take_action(:on_op_assign, [dest] + val[1,2])
                    }
                | primary '.' tIDENTIFIER tOP_ASGN arg
                    {
		      dest = take_action(:on_access, val[0,3])
		      take_action(:on_op_assign, [dest] + val[3,2])
                    }
                | primary '.' tCONSTANT tOP_ASGN arg
                    {
		      dest = take_action(:on_access, val[0,3])
		      take_action(:on_op_assign, [dest] + val[3,2])
                    }
                | primary tCOLON2 tIDENTIFIER tOP_ASGN arg
                    {
		      dest = take_action(:on_access, val[0,3])
		      take_action(:on_op_assign, [dest] + val[3,2])
                    }
                | backref tOP_ASGN arg
                    {
		      backref_error(val[0])
		      nil
                    }
                | arg tDOT2 arg
                    { take_action(:on_range, val) }
                | arg tDOT3 arg
                    { take_action(:on_range, val) }
                | arg '+' arg
                    { take_action(:on_plus_expr, val) }
                | arg '-' arg
                    { take_action(:on_minus_expr, val) }
                | arg '*' arg
                    { take_action(:on_mult_expr, val) }
                | arg '/' arg
                    { take_action(:on_div_expr, val) }
                | arg '%' arg
                    { take_action(:on_mod_expr, val) }
                | arg tPOW arg
                    { take_action(:on_pow_expr, val) }
                | tUPLUS arg
                    { take_action(:on_unaryplus_expr, val) }
                | tUMINUS arg
                    { take_action(:on_unaryminus_expr, val) }
                | arg '|' arg
                    { take_action(:on_or_expr, val) }
                | arg '^' arg
                    { take_action(:on_exor_expr, val) }
                | arg '&' arg
                    { take_action(:on_and_expr, val) }
                | arg tCMP arg
                    { take_action(:on_compare_expr, val) }
                | arg '>' arg
                    { take_action(:on_greaterthan_expr, val) }
                | arg tGEQ arg
                    { take_action(:on_geq_expr, val) }
                | arg '<' arg
                    { take_action(:on_lessthan_expr, val) }
                | arg tLEQ arg
                    { take_action(:on_leq_expr, val) }
                | arg tEQ arg
                    { take_action(:on_eq_expr, val) }
                | arg tEQQ arg
                    { take_action(:on_eqq_expr, val) }
                | arg tNEQ arg
                    { take_action(:on_neq_expr, val) }
                | arg tMATCH arg
                    { take_action(:on_match_expr, val) }
                | arg tNMATCH arg
                    { take_action(:on_notmatch_expr, val) }
                | '!' arg
                    {
                      # Check that val[1] is valid value expr
		      # Make cond of val[1]
		      take_action(:on_exclamation, val)
                    }
                | '~' arg
                    { take_action(:on_invert_expr, val) }
                | arg tLSHFT arg
                    { take_action(:on_lshift_expr, val) }
                | arg tRSHFT arg
                    { take_action(:on_rshift_expr, val) }
                | arg tANDOP arg
                    { take_action(:on_cond_and_expr, val) }
                | arg tOROP arg
                    { take_action(:on_cond_or_expr, val) }
                | kDEFINED opt_nl {@in_defined = true} arg
                    {
		      @in_defined = false
		      take_action(:on_defined, val.indices(0,1,3))
                    }
                | arg '?' arg ':' arg
                    {
                      # Check that val[0] is valid value expr
		      # Make cond of val[0]
		      take_action(:on_if, [nil] + val)
                    }
                | primary
                    { val[0] }

aref_args       : none 
                    { nil }
                | command opt_nl
                    { [[val[0]], nil] }
                | args trailer
                    { [val[0], nil] }
                | args ',' tSTAR arg opt_nl
                    {
		      # Check that val[3] is value expression
		      [val[0], val[3]]
                    }
                | assocs trailer
                    { [val[0], nil] }
                | tSTAR arg opt_nl
                    {
		      # Check that val[1] is value expression
		      [[], val[1]]
                    }

paren_args      : '(' none ')'
		    { nil }
                | '(' call_args opt_nl ')'
		    { val[1] }
                | '(' block_call opt_nl ')'
                    { take_action(:on_args, val[1]) }
                | '(' args ',' block_call opt_nl ')'
                    { take_action(:on_args, val[1].push(val[3])) }

opt_paren_args  : none
                    { nil }
                | paren_args

call_args       : command
		    { take_action(:on_args, val) }
                | args opt_block_arg
                    { take_action(:on_args, 
				  [val[0], nil, nil, nil, nil, nil, val[1]]) }
                | args ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, val[0,2] + ([nil]*2) + val[2,3]) }
                | assocs opt_block_arg
                    { take_action(:on_args, 
				  [nil, nil, val[0], nil, nil, nil, val[1]]) }
                | assocs ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, ([nil]*2) + val) }
                | args ',' assocs opt_block_arg
                    { take_action(:on_args, val[0,3] + ([nil]*3) + val[3]) }
                | args ',' assocs ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, val) }
                | tSTAR arg opt_block_arg
                    { take_action(:on_args, ([nil]*4) + val) }
                | block_arg
                    { take_action(:on_args, ([nil]*6) + val) }

call_args2      : arg ',' args opt_block_arg
                    { take_action(:on_args, 
				  [[val[0], *val[2]], nil, nil, nil, nil, 
				   nil, val[1]]) }
                | arg ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, 
				  [val[0]] + val[1,2] + ([nil]*2) + val[2,3]) }
                | arg ',' args ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, 
				  [[val[0], *val[2]] + ([nil]*2) + val[3,4]]) }
                | assocs opt_block_arg
                    { take_action(:on_args, 
				  [nil, nil, val[0], nil, nil, nil, val[1]]) }
                | assocs ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, ([nil]*2) + val) }
                | arg ',' assocs opt_block_arg
                    { take_action(:on_args, 
				  [[val[0]]] + val[1,2] + ([nil]*3) + val[3]) }
                | arg ',' args ',' assocs opt_block_arg
                    { take_action(:on_args, 
				  [val[0], *val[2]] + val[3,2] + 
				  ([nil]*3) + val[3]) }
                | arg ',' assocs ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, [[val[0]]] + val[1,6]) }
                | arg ',' args ',' assocs ',' tSTAR arg opt_block_arg
                    { take_action(:on_args, [[val[0], *val[2]]] + val[3,6]) }
                | tSTAR arg opt_block_arg
                    { take_action(:on_args, ([nil]*4) + val) }
                | block_arg
                    { take_action(:on_args, ([nil]*6) + val) }

command_args    :
		    {
		      #res = @lexer.cmdarg_stack.snapshot
		      #@lexer.cmdarg_stack.push(1)
		      #res
                    }
                  open_args
                    {
		      #@lexer.cmdarg_stack.pop
		      #@lexer.cmdarg_stack.set_state val[0]
		      val[1]
                    }

open_args       : call_args
                | tLPAREN_ARG  {@lexer.state = EXPR_ENDARG} ')'
                    {
		      rb_warning("#{val[0].inspect} interpreted as method call")
		      raise NotImplementedError
                    }
                | tLPAREN_ARG call_args2 {@lexer.state = EXPR_ENDARG;} ')'
                    {
		      rb_warning("#{val[0].inspect} interpreted as method call")
		      val[1]
                    }

block_arg       : tAMPER arg
                    {
                      # value_expr(val[1])
		      val[1]
                    }

opt_block_arg   : ',' block_arg
                    { val[1] }
                | none
                    { nil }

args            : arg
                    {
                      # value_expr(val[0])
		      val
                    }
                | args ',' arg
                    {
                      # value_expr(val[2])
		      val[0].push(val[2])
                    }

mrhs            : arg
                    {
                      # value_expr(val[0])
		      take_action(:on_mrhs, val)
                    }
                | mrhs_basic
                    { val[0] }

mrhs_basic      : args ',' arg
                    {
                      # value_expr(val[2])
		      take_action(:on_mrhs, [val[0].push(val[2])])
                    }
                | args ',' tSTAR arg
                    {
                      # value_expr(val[3])
		      take_action(:on_mrhs, [val[0], val[3]])
                    }
                | tSTAR arg
                    {
                      # value_expr(val[1])
		      take_action(:on_mrhs, [[], val[1]])
                    }

primary         : literal
                | string
                | tXSTRING
		    { take_action(:on_xstring, val) }
                | tDXSTRING
                | tDREGEXP
                | var_ref
                | backref
                | tFID
                | kBEGIN compstmt rescue opt_else ensure kEND
                    {
		      if val[3] && !val[2]
		        rb_warning("else without rescue is useless")
                      end
		      take_action(:on_begin_end_block, val)
                    }
                | tLPAREN_ARG expr {@lexer.state = EXPR_ENDARG} ')'
                    {
		      rb_warning("#{val[0].inspect} interpreted as method call")
		      val[1]
		    }
                | tLPAREN compstmt ')'
                    { val[1] }
                | primary tCOLON2 tCONSTANT
                    {
		      # print warning!?
		      take_action(:on_access, val)
                    }
                | tCOLON3 cname
                    { take_action(:on_access, [nil] + val) }
                | primary '[' aref_args ']'
                    {
                      # value_expr(val[0])
		      take_action(:on_slice_var, val)
                    }
                | tLBRACK aref_args ']'
                    { take_action(:on_array, val) }
                | tLBRACE assoc_list '}'
                    { take_action(:on_hash, val) }
                | kRETURN
                    {
		      if (!@compile_for_eval && !(@in_def>0) && !@in_single)
		        yyerror("return appeared outside of method") 
		      end
		      take_action(:on_return, val)
                    }
                | kYIELD '(' call_args ')'
                    {
                      #value_expr(val[2])
                      #ret_args(val[2])
		      unless nil == val[2][2]
                        rb_compile_error("block argument should not be given in yield")
                      end
		      take_action(:on_yield, val.indices(0,2))
                    }
                | kYIELD '(' ')'
		    { take_action(:on_yield, val.indices(0)) }
                | kYIELD
		    { take_action(:on_yield, val.indices(0)) }
                | kDEFINED opt_nl '(' {@in_defined = true} expr ')'
                    {
		      @in_defined = false
		      take_action(:on_defined, val.indices(0,1,4))
                    }
                | operation brace_block
                    {
		      raise NotImplementedError
                    }
                | method_call
                | method_call brace_block
                    {
		      if val[0].arguments && val[0].arguments.block
		        rb_compile_error("both block arg and actual block given")
                      end
		      take_action(:on_call_with_block, val)
                    }
                | kIF expr then compstmt if_tail kEND
                    {
                      # value_expr(val[1])
		      take_action(:on_if, val[0,4] + [nil] + val[4,2])
                    }
                | kUNLESS expr then compstmt opt_else kEND
                    {
		      # value_expr(val[1])
		      take_action(:on_unless, val[0,4] + [nil] + val[4,2])
                    }
                | kWHILE            {} #@lexer.cond_stack.push(1)}
                  expr do           {} #@lexer.cond_stack.pop()}
                  compstmt kEND
                    {
                      # value_expr(val[2])
		      take_action(:on_while, val.indices(0,2,3,5,6))
                    }
                | kUNTIL            {} #@lexer.cond_stack.push(1)}
                  expr do           {} #@lexer.cond_stack.pop()} 
                  compstmt kEND
                    {
                      # value_expr(val[2])
		      take_action(:on_until, val.indices(0,2,3,5,6))
                    }
                | kCASE expr opt_terms case_body kEND
                    {
                      # value_expr(val[1])
		      take_action(:on_case, val)
                    }
                | kCASE opt_terms case_body kEND
                    {
		      val[1,0] = [nil]
		      take_action(:on_case, val)
                    }
                | kFOR block_var kIN {} #@lexer.cond_stack.push(1)}
                  expr do            {} #@lexer.cond_stack.pop()}
                  compstmt kEND
                    {
                      #value_expr(val[4])
		      take_action(:on_for, val.indices(0,1,2,4,5,7,8))
                    }
                | kCLASS cname superclass
                    {
		      if (@in_def>0 || @in_single)
		        yyerror("class definition in method body")
		      end
		      @class_nest += 1
                    }
                  compstmt kEND
                    {
		      @class_nest -= 1
		      take_action(:on_class, val.indices(0,1,2,4,5))
		    }
                | kCLASS tLSHFT expr
                    {
		      result, @in_def = @in_def, 0
		      result
                    }
                  term
                    {
		      result, @in_single = @in_single, false
		      @class_nest += 1
		      result
                    }
                  compstmt
                  kEND
                    {
		      @class_nest -= 1
		      @in_def, @in_single = val[3], val[5]
		      take_action(:on_singleton_class, 
				  val.indices(0,1,2,4,6,7))
                    }
                | kMODULE cname
                    {
		      if (@in_def>0 || @in_single)
		        yyerror("module definition in method body")
		      end
		      @class_nest+=1
                    }
                  compstmt
                  kEND
                    {
		      @class_nest -= 1
		      take_action(:on_module, val.indices(0,1,3,4))
		    }
                | kDEF fname
                    {
		      if (@in_def>0 || @in_single)
                        yyerror # ("nested method definition")
		      end
		      @in_def += 1
                    }
                  f_arglist compstmt rescue opt_else ensure kEND
                    {
                      @in_def -= 1
		      body = val[4]
		      if val[5].length > 0 || val[7]
		        body = take_action(:on_begin_end_block, val[4,4])
		      end
		      if val[6] && val[5].length == 0
		        rb_warning("else without rescue is useless")
		      end
		      # I've dropped lots of code related to NOEX_PRIVATE
                      # since I don't (yet) understand it. CHECK!
		      take_action(:on_method, val[0,2] + [val[3], body, val[8]])
                    }
                | kDEF singleton dot_or_colon {@lexer.state = EXPR_FNAME}
                  fname
                    {
                      # value_expr(val[1])
		      result, @in_single = @in_single, true
		      @lexer.state = EXPR_END
		      result
                    }
                  f_arglist compstmt rescue opt_else ensure kEND
                    {
		      body = val[7]
		      if val[8].length > 0 || val[10]
		        body = take_action(:on_begin_end_block, val[7,4])
		      end
		      if val[9] && val[8].length == 0
		        rb_warning("else without rescue is useless")
		      end
		      @in_single = val[5]
		      take_action(:on_class_method,
				  val[0,3] + [val[4],val[6],body, val[11]])
                    }
                | kBREAK
		    { take_action(:on_break, val) }
                | kNEXT
		    { take_action(:on_next, val) }
                | kREDO
		    { take_action(:on_redo, val) }
                | kRETRY
		    { take_action(:on_retry, val) }

then            : term
                | kTHEN
                | term kTHEN

do              : term
                | kDO_COND

if_tail         : opt_else
                | kELSIF expr then
                  compstmt
                  if_tail
                    { take_action(:on_if, val[0,4] + [nil] + val[4,1]) }

opt_else        : none
                | kELSE compstmt
                    { val[1] }

block_var       : lhs
                | mlhs

opt_block_var   : none
		    { take_action(:on_block_var, []) }
                | '|'  # none '|'
		    { take_action(:on_block_var, []) }
                | tOROP
		    { take_action(:on_block_var, []) }
                | '|' block_var '|'
                    { val[1] }

do_block        : kDO_BLOCK
                    {
                      # What? result = dyna_push()
                    }
                  opt_block_var compstmt kEND
                    { take_action(:on_block, val.indices(0, 2, 3, 4)) }
                | tLBRACE_ARG
                    {
                      # What? result = dyna_push()
                    }
                  opt_block_var compstmt '}'
                    { take_action(:on_block, val.indices(0, 2, 3, 4)) }

block_call      : command do_block
                    {
		      rb_compile_error "both block arg and actual block given"
		      # Sometimes valid? Check matx code here!
                    }
                | block_call '.' operation2 opt_paren_args
                    {
                      # val[0] should be value expression
		      take_action(:on_call, val)
                    }
                | block_call tCOLON2 operation2 opt_paren_args
                    {
                      # val[0] should be value expression
		      take_action(:on_call, val)
                    }

method_call     : operation paren_args
		    { take_action(:on_call, [nil, nil, *val]) }
                | primary '.' operation2 opt_paren_args
		    { take_action(:on_call, val) }
                | primary tCOLON2 operation2 paren_args
		    { take_action(:on_call, val) }
                | primary tCOLON2 operation3
		    { take_action(:on_constant, val) }
                | kSUPER paren_args
		    {
		      if (!@compile_for_eval && !(@in_def>0) &&
			  !@in_single && !@in_defined)
		      yyerror("super called outside of method")
		      end
		      take_action(:on_super, val)
                    }
                | kSUPER
                    {
		      if (!@compile_for_eval && !(@in_def>0) &&
			  !@in_single && !@in_defined)
		      yyerror("super called outside of method")
		      end
		      take_action(:on_super, val)
                    }

brace_block     : '{'
                    {
                      # What to do? res = dyna_push()
                    }
                  opt_block_var
                  compstmt '}'
		    { take_action(:on_block, val.indices(0,2,3,4)) }
                | kDO
                    {
                      # What to do? res = dyna_push();
                    }
                  opt_block_var
                  compstmt kEND
		    { take_action(:on_block, val.indices(0,2,3,4)) }

case_body       : kWHEN when_args then
                  compstmt
                  cases
		    { take_action(:on_when, val) }

when_args       : args
		    { [val[0], nil] }
                | args ',' tSTAR arg
                    { [val[0], val[3]] }
                | tSTAR arg
                    { [nil, val[1]] }

cases           : opt_else
                | case_body

exc_list        : none
                | args

exc_var         : tASSOC lhs
		    { val[1] }
                | none

rescue          : kRESCUE exc_list exc_var then
                  compstmt
                  rescue
		    { take_action(:on_rescue, val) }
                | none { [] }

ensure          : none
                | kENSURE compstmt
		    { val[1] }

literal         : numeric
                | symbol
                | tREGEXP

string          : tSTRING
                    { take_action(:on_string, val) }
                | tDSTRING
                    { take_action(:on_dstring, val) }
                | string tSTRING
		    { take_action(:on_strings, val) }
                | string tDSTRING
		    { take_action(:on_strings, val) }

symbol          : tSYMBEG sym
                    {
		      @lexer.state = EXPR_END
		      take_action(:on_symbol, val)
                    }

sym             : fname
                | tIVAR
                | tGVAR
                | tSVAR
                | tCVAR

numeric         : tINTEGER
                | tFLOAT

variable        : tIDENTIFIER
                | tIVAR
                | tGVAR
                | tSVAR
                | tCONSTANT
                | tCVAR
                | kNIL
                   { take_action(:on_nil, val) }
                | kSELF
                   { take_action(:on_self, val) }
                | kTRUE
                   { take_action(:on_true, val) }
                | kFALSE
                   { take_action(:on_false, val) }

                | k__FILE__
                | k__LINE__

var_ref         : variable  # Check that variable is gettable!

backref         : tNTH_REF
                    {
                      VarSpecial[ val[0] ]
                    }
                | tBACK_REF
                    {
                      VarSpecial[ val[0] ]
                    }


superclass      : term
		    { nil }
                | '<' 
                    { @lexer.state = EXPR_BEG }
                  expr term
		    { val[2] }
                | error term 
                    {yyerrok; nil}

f_arglist       : '(' f_args opt_nl ')'
                    {
                        @lexer.state = EXPR_BEG
			val[1]
                    }
                | f_args term
		    { val[0] }

f_args          : f_arg ',' f_optarg ',' f_rest_arg opt_f_block_arg
		    { take_action(:on_formal_args, val) }
                | f_arg ',' f_optarg                opt_f_block_arg
		    { take_action(:on_formal_args, 
				  val[0,3] + [nil, nil] + val[3,1]) }
                | f_arg ',' f_rest_arg opt_f_block_arg
		    { take_action(:on_formal_args, 
				  val[0,2] + [nil, nil] + val[2,2])}
                | f_arg opt_f_block_arg
		    { take_action(:on_formal_args, 
				  val[0,1] + ([nil]*4) + val[1,1]) }
                | f_optarg ',' f_rest_arg opt_f_block_arg
		    { take_action(:on_formal_args, ([nil]*2) + val)}
                | f_optarg opt_f_block_arg
		    { take_action(:on_formal_args, 
				  ([nil]*2) + val[0,1] + ([nil]*2) + val[1,1])
		    }
                | f_rest_arg opt_f_block_arg
		    { take_action(:on_formal_args, ([nil]*4) + val)}
                | f_block_arg
		    { take_action(:on_formal_args, ([nil]*5) + val)}
                | # none 
		    { nil }

f_norm_arg      : tCONSTANT
                    { yyerror("formal argument cannot be a constant") }
                | tIVAR
                    { yyerror("formal argument cannot be a constant") }
                | tGVAR
                    { yyerror("formal argument cannot be a constant") }
                | tSVAR
                    { yyerror("formal argument cannot be a constant") }
                | tCVAR
                    { yyerror("formal argument cannot be a constant") }
                | tIDENTIFIER
                    {
                      # We should check that its local variable and that
		      # the name has been used previously as arg name
		      val[0]
                    }

f_arg           : f_norm_arg
                    { [val[0]] }
                | f_arg ',' f_norm_arg
                    { val[0].push(val[2]) }

f_opt           : tIDENTIFIER '=' arg
                    {
                      # We should check that its local variable and that
		      # the name has been used previously as arg name
		      raise NotImplementedError
                    }

f_optarg        : f_opt
                    { [val[0]] }
                | f_optarg ',' f_opt
                    { val[0].push(val[2]) }

f_rest_arg      : tSTAR tIDENTIFIER
                    {
                      # We should check that its local variable and that
		      # the name has been used previously as arg name
		      val[1]
                    }
                | tSTAR
		    { #what to return here?
		    }

f_block_arg     : tAMPER tIDENTIFIER
                    {
                      # We should check that its local variable and that
		      # the name has been used previously as arg name
		      val[1]
                    }

opt_f_block_arg : ',' f_block_arg
		    { val[1] }
                | none

singleton       : var_ref
                | '(' {@lexer.state = EXPR_BEG;} expr opt_nl ')'
                    {
		      if literal?(val[2])
		        yyerror("can't define single method for literals.")
		      end
		      val[2]
                    }

assoc_list      : none
                | assocs trailer
                    { val[0] }
                | args trailer
                    {
		      if (val[0].length%2 != 0)
		        yyerror("odd number list for Hash")
		      end
		      val[0]
                    }

assocs          : assoc
                    { [val[0]] }
                | assocs ',' assoc
		    { val[0].push(val[2]) }

assoc           : arg tASSOC arg
		    { take_action(:on_assoc, val) }

operation       : tIDENTIFIER
                | tCONSTANT
                | tFID

operation2      : tIDENTIFIER
                | tCONSTANT
                | tFID
                | op

operation3      : tIDENTIFIER
                | tFID
                | op

dot_or_colon    : '.'
                | tCOLON2

opt_terms       : /* none */
                | terms

opt_nl          : /* none */
                | '\n'

trailer         : /* none */
                | '\n'
                | ','

term            : ';' {yyerrok;}
                | '\n'

terms           : term
                | terms ';' {yyerrok;}

none            : /* none */
                  {
                       result = nil;
                  }

end

---- header ----

require 'rulator/lexer'
require 'rulator/actions'

module Ruby

---- inner ----

attr_accessor :debug

# You can optionally supply your own ParseActions object but if not
# a default when is used. It will build the AST for the Ruby program
# being parsed.
def initialize(actor = Ruby::AbstractSyntaxTreeBuilder.new, debug = false)
  @actor, @debug = actor, debug
end

def parse(aString)
  @lexer = Ruby::Lexer.new(aString)

  # We save comment tokens as we go along and then annotate the terms/nodes
  # with the comments. Found prior to and post parsing the term.
  @comments = Array.new 

  # We save reference to last result produced if it has attr_accessors
  # pre_comments and post_comments where we can save the comments surrounding
  # a term.
  @last_result = nil

  # To speed things up we pre-allocate an array to hold the tokens
  # so we don't need to reallocate it every time.
  @token = Array.new(2)

  # Flags and data structures used in Matz parse.y that we also use
  @cond_stack   = 0   # We use an Integer to represent a bit stack
  @cmdarg_stack = 0   # We use an Integer to represent a bit stack
  @class_nest       = 0
  @in_single        = false
  @in_def           = 0
  @compile_for_eval = false
  @cur_mid          = 0
  @in_defined       = false

  # Now parse!
  do_parse
end

# Take an action and annotate its result with comments if we've encountered 
# any since last action.
def take_action(actionId, results)
  if @debug
    puts "#{actionId.inspect}, #{results.inspect}"
  end
  new_result = @actor.send(actionId, results)
  if new_result && new_result.respond_to?(:pre_comments=)
    if @comments.length > 0
      new_result.pre_comments = @comments
      @last_result.post_comments = @comments if @last_result
      @comments = Array.new
    end
    @last_result = new_result
  else
    @last_result = nil
  end
  new_result
end

private

def backref_error(token)
  puts "Backref error on token: #{token.inspect}"
end

# Get the next token from the lexer. If its a comment token we save it
# and get a new one.
def next_token
  tok = nil
  while not tok
    tok = @lexer.next_token
    #puts "Token: #{tok.inspect}"; STDOUT.flush
    break if tok.token_type == false
    case tok.token_type
    when :tComment, :tBlockComment
      @comments.push tok
      tok = nil
    end
  end
  @token[0], @token[1] = tok.token_type, tok
  puts "token: #{@token[1].lexeme.inspect} (state = #{@lexer.state})" if @debug
  @token
end

def error_unless_alias_valid
  yyerror("alias within method") if (@in_def > 0) || @in_single
end


---- footer ----
end # module Ruby
