require 'rockit/parse/combinator_parsers'
include Rockit::Parse
# Turn a parser for tokens into a lexer
#
module Rockit
  module Parse
    class LexerError < Exception
      def initialize(message, position)
	super(message)
	@position = position
      end
    end

    class Lexer
      def initialize(tokenParser, *streamSpecifyingObjects)
	@parser = tokenParser
	@pstate = OptimizedParseState.new(*streamSpecifyingObjects)
      end

      # Note that the state of the lexer is *NOT* the same as the parser state.
      def state=(newState)
	@pstate.set_attribute(:lexer_state, newState)
      end

      def state
	@pstate.get_attribute(:lexer_state)
      end

      def position; @pstate.position; end

      def next_token
	# TODO: Fix this; its too much low-level fiddling here. Should be in
	# the parser or the parse state.
	@pstate.split
	if @parser.match_and_commit(@pstate)
	  @pstate.pop_result
	else
	  @pstate.back
	  error = @parser.analyze_error(@pstate)
	  raise error if error
	end
      end

      # Create a subclass with some parameters specified. Typically only
      # one parameter: the parser for tokens.
      def Lexer.new_lexer_class(*args)
	cls = Class.new(Lexer)
	define_args_merging_method(cls, args)
	# Make sure the args merging method is called in initialize
	cls.class_eval %{
          def initialize(*a)
	    super(*merge_with_prespecified_args(a))
          end
        }
	cls
      end

      def Lexer.define_args_merging_method(klass, args)
	class << klass; public :define_method; end
	klass.define_method(:merge_with_prespecified_args) {|rest| args + rest}
	class << klass; protected :define_method; end # CHECK: private?
      end
    end
  end # Parse
end # Rockit
