require 'format_doc'

module Rockit
  module PrettyPrint
    class Writer
      def write(anObjectOrFDoc)
	if FormattedDoc === anObjectOrFDoc
	  write_fdoc(anObjectOrFDoc)
	else
	  write_fdoc(anObjectOrFDoc.to_formatted_doc)
	end
      end

      private

      def write_fdoc(aFDoc)
	raise NotImplementedError
      end
    end

    class PlainTextWriter < Writer
      def initialize(options = {})
	@width = options[:width] ||80
	@column = 0
	@indent = options[:indent] || "  "
	# Stack of start columns for CompoundFDoc's. The last value is the
	# start column of the CompoundFDoc being written.
	@start_stack = Array.new
      end

      # Write it for Array:
      #   hor("[", list(self), "]")

      def write_fdoc(aFDoc)
	@str = ""
	recursive_write_fdoc(aFDoc)
	@str
      end

      def in_compound
	@start_stack.push @column
	yield
	@start_stack.pop
      end

      def space_left?(aString)
	aString.length < (@width - @column)
      end

      def next_line
	@str << "\n"
	@column = @start_stack.last
	@str << " " * @column
      end

      def write_string(aString)
	unless space_left?(aString)
	  puts "No space left for #{aString.inspect} (width = #{@width}, col = #{@column}, st = #{@start_stack.inspect})"
	  next_line 
	end
	@str << aString
	@column += aString.length
      end

      def recursive_write_fdoc(aFDoc)
	case aFDoc
	when Text
	  write_string(aFDoc.string)
	when Besides
	  sep = aFDoc.separator
	  in_compound do
	    aFDoc.each_with_index do |d, i| 
	      recursive_write_fdoc(d)
	      write_string(sep) unless i == (aFDoc.length - 1)
	    end
	  end
	when Vertical
	  in_compound do
	    aFDoc.each do |d| 
	      recursive_write_fdoc(d)
	      next_line
	    end
	  end
	end
      end
    end
  end
end
