require 'rubyunit'

require 'rockit/parse/char_set'
include Rockit::Parse

class CharSetTest < TestCase
  def test_01_init
    cs = CharSet.new
    assert_kind_of(CharSet, cs)
    assert_equals(0, cs.length)
  end

  def test_02_length
    cs = CharSet.new "abcc"
    assert_equals(3, cs.length)
  end

  def test_03_include_char?
    cs = CharSet.new "abcc"
    "abc".each_byte do |char|
      assert(cs.include_char?(char))
    end
    assert_equals(false, cs.include_char?(?d))
    assert_equals(false, cs.include_char?(?1))
  end

  def test_04_to_string
    cs = CharSet.new "abcc"
    assert_equals("abc", cs.to_string)
  end

  def test_05_invert
    cs = CharSet.new "abcc"
    ics = cs.invert
    for char in 0..0xff
      assert_equals(!"abc".include?(char), ics.include_char?(char))
    end
  end

  def test_06_equality
    cs = CharSet.new("aggccdd")
    assert(cs == CharSet.new("dgac"))
  end

  def assert_set(string, set)
    assert_equals(CharSet.new(string), set)
  end  

  def test_07_merge
    assert_set("ab", CharSet.new("a").merge(CharSet.new("b")))
    assert_set("abd1", CharSet.new("ad").merge(CharSet.new("b1")))
  end

  def test_08_epsilon_init
    cs = CharSet.new("a")
    assert_equals(false, cs.include_epsilon?)

    cs = CharSet.new("a", true)
    assert_equals(true, cs.include_epsilon?)
  end

  def test_09_epsilon_merging
    cs, cs2 = CharSet.new("a", true), CharSet.new("b")
    assert_equals(true, (csm = cs.merge(cs2)).include_epsilon?)
    assert_equals("ab", csm.to_string)

    cs, cs2 = CharSet.new("a"), CharSet.new("b", true)
    assert_equals(true, (csm = cs.merge(cs2)).include_epsilon?)
    assert_equals("ab", csm.to_string)

    cs, cs2 = CharSet.new("a"), CharSet.new("b")
    assert_equals(false, (csm = cs.merge(cs2)).include_epsilon?)
    assert_equals("ab", csm.to_string)
  end

  def test_10_add_and_remove_epsilon
    cs = CharSet.new("a")
    cs.add_epsilon
    assert_equals(true, cs.include_epsilon?)

    cs.remove_epsilon
    assert_equals(false, cs.include_epsilon?)

    cs.add_epsilon
    assert_equals(true, cs.include_epsilon?)
  end

  def test_11_epsilon_invert
    cs = CharSet.new("a", true)
    csi = cs.invert
    assert_equals(true, csi.include_epsilon?)
  end
end

