require 'rubyunit'

require 'rockit/parse/result_mapping'
include Rockit::Parse

class ResultMappingFunctionTest < RUNIT::TestCase
  def test_01_initialize
    rmf = ResultMappingFunction.new
    assert_kind_of(ResultMappingFunction, rmf)
    assert_equals("a", rmf.call("a"))
  end

  def test_02_tointfunction
    ti = ToIntFunction.new
    assert_kind_of(ResultMappingFunction, ti)
    assert_kind_of(ToIntFunction, ti)

    assert_equals(0, ti.call("0"))
    assert_equals(11, ti.call("11"))
    assert_equals(22, ti.call("22"))
    assert_equals(99, ti.call("99"))
    assert_equals(100, ti.call("100"))
    assert_equals(1234, ti.call("+1234"))
    assert_equals(-98765, ti.call("-98765"))

    assert_exception(ArgumentError) {ti.call("a")}
    assert_exception(ArgumentError) {ti.call("1.5")}
    assert_exception(ArgumentError) {ti.call("10e12")}
    assert_exception(ArgumentError) {ti.call("3.87E-45")}
  end

  def test_03_proc_wrapper
    pf1 = ProcFunction.new proc{|x| x}
    assert_kind_of(ResultMappingFunction, pf1)
    assert_kind_of(ProcFunction, pf1)
    assert_equals(1, pf1.call(1))

    pf2 = ProcFunction.new proc{|res, state| [res, state]}
    assert_kind_of(ResultMappingFunction, pf2)
    assert_kind_of(ProcFunction, pf2)
    assert_equals([1,2], pf2.call(1,2))
  end

  def test_04_hextointfunction
    ti = HexToIntFunction.new
    assert_kind_of(ResultMappingFunction, ti)
    assert_kind_of(HexToIntFunction, ti)

    assert_equals(0,       ti.call("0"))
    assert_equals(0x11,    ti.call("11"))
    assert_equals(0x2a,    ti.call("2a"))
    assert_equals(0xff,    ti.call("ff"))
    assert_equals(0x100,   ti.call("100"))
    assert_equals(0x123e5, ti.call("123e5"))

    assert_exception(ArgumentError) {ti.call("1.5")}
    assert_exception(ArgumentError) {ti.call("3.87E-45")}
  end

  def test_05_octaltointfunction
    ti = OctalToIntFunction.new
    assert_kind_of(ResultMappingFunction, ti)
    assert_kind_of(OctalToIntFunction, ti)

    assert_equals(0,         ti.call("0"))
    assert_equals(011,       ti.call("11"))
    assert_equals(027,       ti.call("27"))
    assert_equals(001234567, ti.call("01234567"))

    assert_exception(ArgumentError) {ti.call("1.5")}
    assert_exception(ArgumentError) {ti.call("3.87E-45")}
  end

  def test_06_binarytointfunction
    ti = BinaryToIntFunction.new
    assert_kind_of(ResultMappingFunction, ti)
    assert_kind_of(BinaryToIntFunction, ti)

    assert_equals(0b0,    ti.call("0"))
    assert_equals(0b1,    ti.call("1"))
    assert_equals(0b0101, ti.call("0101"))
    assert_equals(0b1111, ti.call("1111"))

    assert_exception(ArgumentError) {ti.call("1.5")}
    assert_exception(ArgumentError) {ti.call("3.87E-45")}
  end

  def test_06_tofloatfunction
    ti = ToFloatFunction.new
    assert_kind_of(ResultMappingFunction, ti)
    assert_kind_of(ToFloatFunction, ti)

    assert_equals(1.0,    ti.call("1.0"))
    assert_equals(-2.2_3,    ti.call("-2.2_3"))

    assert_exception(ArgumentError) {ti.call("e")}
  end
end

RUNIT::CUI::TestRunner.new.run(ResultMappingFunctionTest.suite) if $0 == __FILE__

